// Description: Compute exact discretization state-space matrices and their Jacobians.
// Outputs Jacobians with respect to the parameters: R, K (assuming K_E = K_T = K).

#include "constants.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

float compute_f(int s, float K, float a, float b, float alpha, float beta, float t);
float compute_fD(int s, float K, float KD, float a, float aD, float b, float bD, float alpha, float beta, float t);
float comp_lin(float k, float kD, float w, float wD, float g, float gD);

static void comp_ss_jacobi(float **Ad, float **Bd, float **AdR, float **BdR, float **AdK, float **BdK, float R, float K, float T)
{
    int i;
    
    // Model parameters
    float L = L_; // H
    float J = J_; // kg m^2
    //float R = 2.74; // Ohm
    float K_T = K; // Nm/A
    float K_E = K; // V/(rad/s)
    
    //float T = 1e-3; // s
    
    // Butterworth filter cut-off frequency
    float wc = 2*M_PI*100;
    
    float d, l1, l2, e1, e2;
    float lx, ex, cosx, sinx;
    float w;
    float **AdD, **BdD;
    float l1D, l2D, e1D, e2D;
    float wD;
    
    float k[20], kD[20];
    
	k[0]  = 1.0F;
    k[1]  = K_E/L;
	k[2]  = wc;
	k[3]  = -(wc*K_E)/L;
	k[4]  = (float)pow(wc,2);
	k[5]  = -((float)pow(wc,2)*K_E)/L;
	k[6]  = K_T/J;
	k[7]  = 1.0F;
	k[8]  = K_T/J;
	k[9]  = 1.0F;
	k[10] = J/(K_T*L);
    k[11] = wc/L;
	k[12] = (float)pow(wc,2)/L;
	k[13] = 1.0F/L;
	k[14] = K_T/(J*L);
	k[15] = K_E/(K_T*L);
	k[16] = (wc*K_E)/(J*L);
	k[17] = ((float)pow(wc,2)*K_E)/(J*L);
	k[18] = -1.0F/J;
	k[19] = -1.0F/J;
	
    for (i = 0; i < 20; i++) {
        kD[i] = 0.0F;
    }
    
    // Eigenvalues and related sizes
    d = (float)sqrt((float)pow(J,2)*(float)pow(R,2) - 4*J*L*K_T*K_E);
    
    l1 = (-J*R - d)/(2*J*L);
    l2 = (-J*R + d)/(2*J*L);
    e1 = (float)exp(l1*T);
    e2 = (float)exp(l2*T);
    
    lx = wc*(float)sqrt(2)/2;
    ex = (float)exp(-lx*T);
    cosx = (float)cos(lx*T);
    sinx = (float)sin(lx*T);
    
    w = l1 - l2;
    
    // State matrix (6 x 6), row wise
    Ad[1][1] = k[0]*(l1*e1 - l2*e2)/w;
    Ad[1][2] = 0.0;
    Ad[1][3] = 0.0;
    Ad[1][4] = k[1]*(e2 - e1)/w;
    Ad[1][5] = 0.0;
    
    Ad[2][1] = compute_f(3, k[2], l1, l2, -lx, lx, T);
    Ad[2][2] = ( cosx - sinx )*ex;
    Ad[2][3] = -wc*( 1/lx*sinx )*ex;
    Ad[2][4] = compute_f(2, k[3], l1, l2, -lx, lx, T);
    Ad[2][5] = 0.0;
    
    Ad[3][1] = compute_f(2, k[4], l1, l2, -lx, lx, T);
    Ad[3][2] = -Ad[2][3];
    Ad[3][3] = ( sinx + cosx )*ex;
    Ad[3][4] = compute_f(1, k[5], l1, l2, -lx, lx, T);
    Ad[3][5] = 0.0;
    
    Ad[4][1] = k[6]*(e1 - e2)/w;
    Ad[4][2] = 0.0;
    Ad[4][3] = 0.0;
    Ad[4][4] = k[7]*(l1*e2 - l2*e1)/w;
    Ad[4][5] = 0.0;
    
    Ad[5][1] = k[8]*((1.0F/l1)*(e1 - 1.0F) - (1.0F/l2)*(e2 - 1.0F))/w;
    Ad[5][2] = 0.0;
    Ad[5][3] = 0.0;
    Ad[5][4] = k[9]*((l1/l2)*(e2 - 1.0F) - (l2/l1)*(e1 - 1.0F))/w;
    Ad[5][5] = 1.0;
    
    Ad[6][1] = 0.0;
    Ad[6][2] = 0.0;
    Ad[6][3] = 0.0;
    Ad[6][4] = 0.0;
    Ad[6][5] = 0.0;
    
    // Input matrix (6 x 1), column wise
    // Known input (voltage)
    Bd[1][1] = k[10]*Ad[4][1];
    Bd[2][1] = compute_f(2, k[11], l1, l2, -lx, lx, T);
    Bd[3][1] = compute_f(1, k[12], l1, l2, -lx, lx, T);
    Bd[4][1] = k[13]*Ad[5][1];
    Bd[5][1] = k[14]*(1/(float)pow(l1,2)*(e1 - 1) - 1/(float)pow(l2,2)*(e2 - 1) + (1/l2 - 1/l1)*T)/w;
    Bd[6][1] = 0.0;
    // Unknown input (torque load)
    Ad[1][6] = k[15]*Ad[5][1];
    Ad[2][6] = compute_f(1, k[16], l1, l2, -lx, lx, T);
    Ad[3][6] = compute_f(0, k[17], l1, l2, -lx, lx, T);
    Ad[4][6] = k[18]*Ad[5][4];
    Ad[5][6] = k[19]*( l1/(float)pow(l2,2)*(e2 - 1) - l2/(float)pow(l1,2)*(e1 - 1) + (l2/l1 - l1/l2)*T)/w;
    Ad[6][6] = 1.0F;
    
    // Linearization (Jacobians)
    for (i = 1; i <= 2; i++) {
        
        switch (i) {
            case 1: { // Resistance - R
                l1D = -(1.0F + (J*R)/d)/(2*L);
                l2D = -(1.0F - (J*R)/d)/(2*L);
                
                AdD = AdR;
                BdD = BdR;
                break;
            }
            case 2: { // Speed/torque/electromagnetic constant - K_E = K_T = K
                l1D = (2*K)/d;
                l2D = -(2*K)/d;
                
                kD[1] = 1.0F/L;
                kD[3] = -wc/L;
                kD[5] = -(float)pow(wc,2)/L;
                kD[6] = 1.0F/J;
                kD[8] = 1.0F/J;
                kD[10] = -J/((float)pow(K,2)*L);
                kD[14] = 1.0F/(J*L);
                kD[16] = wc/(J*L);
                kD[17] = (float)pow(wc,2)/(J*L);
                
                AdD = AdK;
                BdD = BdK;
                break;
            }
        }
        
        e1D = l1D*T*e1;
        e2D = l2D*T*e2;
        wD = l1D - l2D;
        
        // Linearization of state matrix with respect to ith parameter
        AdD[1][1] = comp_lin(k[0], kD[0], w, wD, (l1*e1 - l2*e2), (l1D*e1 + l1*e1D - l2D*e2 - l2*e2D));
        AdD[1][2] = 0;
        AdD[1][3] = 0;
        AdD[1][4] = comp_lin(k[1], kD[1], w, wD, (e2 - e1), (e2D - e1D));
        AdD[1][5] = 0;
        
        AdD[2][1] = compute_fD(3, k[2], kD[2], l1, l1D, l2, l2D, -lx, lx, T);
        AdD[2][2] = 0;
        AdD[2][3] = 0;
        AdD[2][4] = compute_fD(2, k[3], kD[3], l1, l1D, l2, l2D, -lx, lx, T);
        AdD[2][5] = 0;
        
        AdD[3][1] = compute_fD(2, k[4], kD[4], l1, l1D, l2, l2D, -lx, lx, T);
        AdD[3][2] = 0;
        AdD[3][3] = 0;
        AdD[3][4] = compute_fD(1, k[5], kD[5], l1, l1D, l2, l2D, -lx, lx, T);
        AdD[3][5] = 0;
        
        AdD[4][1] = comp_lin(k[6], kD[6], w, wD, (e1 - e2), (e1D - e2D));
        AdD[4][2] = 0;
        AdD[4][3] = 0;
        AdD[4][4] = comp_lin(k[7], kD[7], w, wD, (l1*e2 - l2*e1), (l1D*e2 + l1*e2D - l2D*e1 - l2*e1D));
        AdD[4][5] = 0;
        
        AdD[5][1] = comp_lin(k[8], kD[8], w, wD, ((e1 - 1)/l1 - (e2 - 1)/l2), (e1D/l1 - (l1D*(e1 - 1))/(float)pow(l1,2) - e2D/l2 + (l2D*(e2 - 1))/(float)pow(l2,2)));
        AdD[5][2] = 0;
        AdD[5][3] = 0;
        AdD[5][4] = comp_lin(k[9], kD[9], w, wD, ((l1*(e2 - 1))/l2 - (l2*(e1 - 1))/l1),
                    ((l1D*l2 - l1*l2D)/(float)pow(l2,2)*(e2 - 1) + (l1*e2D)/l2 - (l2D*l1 - l2*l1D)/(float)pow(l1,2)*(e1 - 1) - (l2*e1D)/l1));
        AdD[5][5] = 0;
        
        AdD[6][1] = 0.0;
        AdD[6][2] = 0.0;
        AdD[6][3] = 0.0;
        AdD[6][4] = 0.0;
        AdD[6][5] = 0.0;
        
        // Input matrix
        // Known input (voltage)
        BdD[1][1] = kD[10]*Ad[4][1] + k[10]*AdD[4][1];
        BdD[2][1] = compute_fD(2, k[11], kD[11], l1, l1D, l2, l2D, -lx, lx, T);
        BdD[3][1] = compute_fD(1, k[12], kD[12], l1, l1D, l2, l2D, -lx, lx, T);
        BdD[4][1] = kD[13]*Ad[5][1] + k[13]*AdD[5][1];
        BdD[5][1] = comp_lin(k[14], kD[14], w, wD, ((e1 - 1)/(float)pow(l1,2) - (e2 - 1)/(float)pow(l2,2) + (1/l2 - 1/l1)*T),
                    ( -(2*l1*l1D)/(float)pow(l1,4)*(e1 - 1) + e1D/(float)pow(l1,2) + (2*l2*l2D)/(float)pow(l2,4)*(e2 - 1) - e2D/(float)pow(l2,2)) + (l1D/(float)pow(l1,2) - l2D/(float)pow(l2,2))*T );
        BdD[6][1] = 0.0;
        // Unknown input (torque load)
        AdD[1][6] = kD[15]*Ad[5][1] + k[15]*AdD[5][1];
        AdD[2][6] = compute_fD(1, k[16], kD[16], l1, l1D, l2, l2D, -lx, lx, T);
        AdD[3][6] = compute_fD(0, k[17], kD[17], l1, l1D, l2, l2D, -lx, lx, T);
        AdD[4][6] = kD[18]*Ad[5][4] + k[18]*AdD[5][4];
        AdD[5][6] = comp_lin(k[19], kD[19], w, wD, (l1/(float)pow(l2,2)*(e2 - 1) - l2/(float)pow(l1,2)*(e1 - 1) + (l2/l1 - l1/l2)*T),
                    ( (l1D*(float)pow(l2,2) - l1*2*l2*l2D)/(float)pow(l2,4)*(e2 - 1) + (l1*e2D)/(float)pow(l2,2) - (l2D*(float)pow(l1,2) - l2*2*l1*l1D)/(float)pow(l1,4)*(e1 - 1) - (l2*e1D)/(float)pow(l1,2) + ((l2D*l1 - l2*l1D)/(float)pow(l1,2) - (l1D*l2 - l1*l2D)/(float)pow(l2,2))*T ));
        AdD[6][6] = 0.0;
        
    }
}


float compute_f(int s, float K, float a, float b, float alpha, float beta, float t)
{
    float u, v, f;
    
    float ea = (float)exp(a*t);
    float eb = (float)exp(b*t);
    float ex = (float)exp(alpha*t);
    float sinx = (float)sin(beta*t);
    float cosx = (float)cos(beta*t);
    
    float d1 = (2*alpha*a - (float)pow(a,2) - (float)pow(alpha,2) - (float)pow(beta,2));
    float d2 = (2*alpha*b - (float)pow(b,2) - (float)pow(alpha,2) - (float)pow(beta,2));
    
    float A1 = K/(d1*(b - a));
    float A2 = K/(d2*(a - b));
    float M = (K*(a + b - 2*alpha))/(d1*d2);
    float N = (K*(3*(float)pow(alpha,2) - (float)pow(beta,2) - 2*alpha*(a + b) + a*b))/(d1*d2);
    
    switch (s) {
        case 0: {
            u = (N + alpha*M)/beta;
            v = M;
            f = A1*(1.0F/a)*(ea - 1.0F) + A2*(1.0F/b)*(eb - 1.0F) + 
                    ( u*((alpha*sinx - beta*cosx)*ex + beta) + v*((alpha*cosx + beta*sinx)*ex - alpha) )/((float)pow(alpha,2) + (float)pow(beta,2));
            break;
        }
        case 1: {
            u = (N + alpha*M)/beta;
            v = M;
            f = A1*ea + A2*eb + ( u*sinx + v*cosx )*ex;
            break;
        }
        case 2: {
            u = ( alpha*N + ((float)pow(alpha,2) - (float)pow(beta,2))*M )/beta;
            v = N + 2*alpha*M;
            f = a*A1*ea + b*A2*eb + ( u*sinx + v*cosx )*ex;
            break;
        }
        case 3: {
            u = ( ((float)pow(alpha,2) - (float)pow(beta,2))*N + ((float)pow(alpha,3) - 3*alpha*(float)pow(beta,2))*M )/beta;
            v = 2*alpha*N + (3*(float)pow(alpha,2) - (float)pow(beta,2))*M;
            f = (float)pow(a,2)*A1*ea + (float)pow(b,2)*A2*eb + ( u*sinx + v*cosx )*ex;
            break;
        }
    }
    
    return f;
}


float compute_fD(int s, float K, float KD, float a, float aD, float b, float bD, float alpha, float beta, float t)
{
    float uD, vD, fD;
    
    float ea = (float)exp(a*t);
    float eb = (float)exp(b*t);
    float ex = (float)exp(alpha*t);
    float sinx = (float)sin(beta*t);
    float cosx = (float)cos(beta*t);
    
    float d1 = (2*alpha*a - (float)pow(a,2) - (float)pow(alpha,2) - (float)pow(beta,2));
    float d2 = (2*alpha*b - (float)pow(b,2) - (float)pow(alpha,2) - (float)pow(beta,2));
    float d1D = 2*alpha*aD - 2*a*aD;
    float d2D = 2*alpha*bD - 2*b*bD;
    
    float A1 = K/(d1*(b - a));
    float A2 = K/(d2*(a - b));
    
    float A1D = (KD*d1*(b - a) - K*(d1D*(b - a) + d1*(bD - aD)))/(float)pow(d1*(b - a),2);
    float A2D = (KD*d2*(a - b) - K*(d2D*(a - b) + d2*(aD - bD)))/(float)pow(d2*(b - a),2);
    float MD = ((KD*(a + b - 2*alpha) + K*(aD + bD))*d1*d2 - K*(a + b - 2*alpha)*(d1D*d2 + d1*d2D))/(float)pow(d1*d2,2);
    float ND = ((KD*(3*(float)pow(alpha,2) - (float)pow(beta,2) - 2*alpha*(a + b) + a*b) + K*(aD*b + a*bD - 2*alpha*(aD + bD)))*d1*d2 -
                    K*(3*(float)pow(alpha,2) - (float)pow(beta,2) - 2*alpha*(a + b) + a*b)*(d1D*d2 + d1*d2D))/(float)pow(d1*d2,2);
    
    switch (s) {
        case 0: {
            uD = (ND + alpha*MD)/beta;
            vD = MD;
            fD = (A1D*(ea - 1.0F))/a + (A1*(aD*t*ea*a - (ea - 1.0F)*aD))/(float)pow(a,2) + 
                    (A2D*(eb - 1.0F))/b + (A2*(bD*t*eb*b - (eb - 1.0F)*bD))/(float)pow(b,2) +
                        ( uD*((alpha*sinx - beta*cosx)*ex + beta) +
                            vD*((alpha*cosx + beta*sinx)*ex - alpha) )/((float)pow(alpha,2) + (float)pow(beta,2));
            break;
        }
        case 1: {
            uD = (ND + alpha*MD)/beta;
            vD = MD;
            fD = (A1D + A1*aD*t)*ea + (A2D + A2*bD*t)*eb + ( uD*sinx + vD*cosx )*ex;
            break;
        }
        case 2: {
            uD = ( alpha*ND + ((float)pow(alpha,2) - (float)pow(beta,2))*MD )/beta;
            vD = ND + 2*alpha*MD;
            fD = (aD*A1 + a*A1D + a*A1*aD*t)*ea + (bD*A2 + b*A2D + b*A2*bD*t)*eb +
                    ( uD*sinx + vD*cosx )*ex;
            break;
        }
        case 3: {
            uD = ( ((float)pow(alpha,2) - (float)pow(beta,2))*ND + ((float)pow(alpha,3) - 3*alpha*(float)pow(beta,2))*MD )/beta;
            vD = 2*alpha*ND + (3*(float)pow(alpha,2) - (float)pow(beta,2))*MD;
            fD = (2*a*aD*A1 + (float)pow(a,2)*A1D + (float)pow(a,2)*A1*aD*t)*ea +
                    (2*b*bD*A2 + (float)pow(b,2)*A2D + (float)pow(b,2)*A2*bD*t)*eb +
                        ( uD*sinx + vD*cosx )*ex;
            break;
        }
    }
    
    return fD;
}


float comp_lin(float k, float kD, float w, float wD, float g, float gD)
{
    float l = k*(gD*w - g*wD)/(float)pow(w,2) + (kD*g)/w;
    
    return l;
}
