% Simulink initialization for DC motor parameter estimation.

disp('Init...')

if (exist('Wref', 'var') == 0)
    error('Need to set Wref (angular velocity reference)')
end
if (exist('Weight', 'var') == 0)
    error('Need to set Weight (weight motor is loaded with)')
end

% Initialize thermal model
thermal_model 

% Data for DC motor (Maxon RE 36.118800)
L = 0.487e-3; %mH -> H
R = 2.74; %Ohm
K_T = 56.6e-3; %mNm/A -> Nm/A
K_E = K_T; %V/(rad/s)
Jm = 6.78e-6; %g cm^2 -> kg m^2
K = K_E;

Js1 = 0.85e-6; %kg m^2
Js2 = 1.1e-6; %kg m^2
Js3 = 1.6e-6; %kg m^2

J = Js3 + Jm;
%J = Jm;

% Bias
v_bias = 0.015; % Volt
i_bias = 0.000; % Ampere

% Input/output scaling factors
K_amp = 1/3; % Volt/Volt
K_curr = 5/4; % Ampere/Volt
K_enc = (2*pi)/(2000); % rad/count

% Controller gains
if 0
    K_p = 10; % Postion controller
else
    f_0 = 27; % Hz
    omega_0 = 2*pi*f_0; % Bandwidth
    zeta = sqrt(2); % Damping
    K_p = omega_0^2*J; % PD controller
    K_d = 2*zeta*omega_0*J;
end

% Gain-scheduler
if 1
    alpha = 0.80*142.256;
    beta  = 102.330;
    gamma = 334.304;
else
    alpha = 3270.2;
    beta  = 2009.8;
    gamma = 8779.7;
end
% GS adapt
mu_gs = 100.0;

% Initial values: R, K, ib
x0RKib = [0, 0, 0, 0, 0, 0, 10, 56.6e-3, 0.001];

q0RKib = 1e-21*ones(1,9);
q0RKib(6) = 1.0e-09; % T_L
q0RKib(7) = 1.0e-07; % R
%q0RKib(7) = 1.0e-05; % R
q0RKib(8) = 1.0e-12; % K
q0RKib(9) = 1.0e-15; % i_b

q0RKib = (Ts/5.0e-4)*q0RKib; % Scaling variances for different sampling time

% Initial values: R, ib
x0Rib = [0, 0, 0, 0, 0, 0, 10, 0.001];

switch 1
    case 1 % Comparison
        q0Rib = 1.0e-21*ones(1,8);
        q0Rib(6) = 1.0e-06; % T_L
        q0Rib(7) = 1.0e-00; % R
        q0Rib(8) = 1.0e-15; % i_b
    case 2 % Estimation for nonlin.reg.
        q0Rib = 1.0e-21*ones(1,8);
        q0Rib(6) = 1.0e-09; % T_L
        q0Rib(7) = 1.0e-07; % R
        q0Rib(8) = 1.0e-15; % i_b
end

q0Rib = (Ts/5.0e-4)*q0Rib; % Scaling variances for different sampling time

% Butterworth filter
f_c = 100; % Hz
omega_c = 2*pi*f_c; % rad/s
[Ab, Bb, Cb, Db] = butter(2,omega_c,'s');
sys = ss(Ab, Bb, Cb, Db);
sysd = c2d(sys, Ts, 'zoh');

% Gradient method/LMS scaling constant
mu = 0.02;

if 1
    % Reference trajectory 1 (lift weight off the floor)
    s0 = dir*0.0;
    s1 = dir*0.3;
    %s1 = dir*2.05*pi*7.5e-3;
    
    t0 = 0:Ts:(1 - Ts);
    th0 = zeros(1,length(t0));
    om0 = zeros(1,length(t0));
    
    [t1, th1, om1] = trajectory_generator(s0, s1, 0.1, 10, Ts);
    
    time3 = 0:Ts:(t0(end) + t1(end) + 1*Ts);
    signal3 = [th0 th1];
    time4 = time3;
    signal4 = [om0 om1];
    
    simin3 = [time3' signal3'];
    simin4 = [time4' signal4'];
end

if 1
    % Reference trajectory 2 (run up/down, full length)
    r = 7.5e-3; % Spindle radius
    
    s0 = dir*0.0;
    s1 = dir*0.05;
    s2 = dir*0.55;
    
    t0 = 0:Ts:(1 - Ts); % wait a second
    Th0 = zeros(1,length(t0));
    Om0 = zeros(1,length(t0));
    
    [t1, Th1, Om1] = trajectory_generator(s0, s1, 0.1, 10, Ts); % clear the floor
    
    t2 = 0:Ts:(0.5 - Ts);
    Th2 = (s1/r)*ones(1,length(t2));
    Om2 = zeros(1,length(t2));
    
    % Excite the system for a little while
    t3 = 0:Ts:(3 - Ts);
    A = dir*0.04;
    Th3 = (A*sin(2*pi*t3 + 3*pi/2) + s1 + A)/r;
    Om3 = (A*2*pi*cos(2*pi*t3 + 3*pi/2))/r;
    
    [t4, Th4, Om4] = trajectory_generator(s1, s2, 0.1, Wref, Ts); % up
    
    t5 = 0:Ts:(0.5 - Ts);
    Th5 = (s2/r)*ones(1,length(t2));
    Om5 = zeros(1,length(t2));
    
    % Excite the system for a little while
    t6 = 0:Ts:(3 - Ts);
    A = dir*0.04;
    Th6 = (A*sin(2*pi*t6 + pi/2) + s2 - A)/r;
    Om6 = (A*2*pi*cos(2*pi*t6 + pi/2))/r;
    
    [t7, Th7, Om7] = trajectory_generator(s2, s1, 0.1, Wref, Ts); % down
    
    [t8, Th8, Om8] = trajectory_generator(s1, s0, 0.1, 10, Ts); % back to floor
    
    % Total time
    time = 0:Ts:(2*t0(end) + t1(end) + 3*t2(end) + t3(end) + t4(end) + 2*t5(end) + t6(end) + t7(end) + t8(end) + 12*Ts);
    
    % Markers for start/end, constant velocity
    T0 = t0(end); % Commit (note temperature at this point)
    
    T1 = t0(end) + t1(end) + 2*t2(end) + t3(end) + 4*Ts + 0.1; % Start const. vel up
    T2 = t0(end) + t1(end) + 2*t2(end) + t3(end) + t4(end) + 5*Ts - 0.1; % End const. vel up
    T3 = t0(end) + t1(end) + 2*t2(end) + t3(end) + t4(end) + 2*t5(end) + t6(end) + 8*Ts + 0.1; % Start const. vel down
    T4 = t0(end) + t1(end) + 2*t2(end) + t3(end) + t4(end) + 2*t5(end) + t6(end) + t7(end) + 9*Ts - 0.1; % End const. vel down
    
    T5 = t0(end) + t1(end) + 3*t2(end) + t3(end) + t4(end) + 2*t5(end) + t6(end) + t7(end) + t8(end) + 11*Ts; % Release (note temperature at this point)
    
    signal_pos = [Th0 Th1 Th2 Th3 Th2 Th4 Th5 Th6 Th5 Th7 Th2 Th8 Th0];
    signal_vel = [Om0 Om1 Om2 Om3 Om2 Om4 Om5 Om6 Om5 Om7 Om2 Om8 Om0];
    
    simin_pos = [time' signal_pos'];
    simin_vel = [time' signal_vel'];
    
    trajectory_filename = ['trajectory-' num2str(Wref) '-rad-s'];
    if ~(exist(['data\' trajectory_filename '.mat'], 'file') == 2)
        save(['data\' trajectory_filename], 's0', 's1', 's2', ...
            'T0', 'T1', 'T2', 'T3', 'T4', 'T5', 'signal_pos', 'simin_vel', 'time');
    end
end
