#define S_FUNCTION_NAME  joint_ekf_R_ib
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"
#include "math.h"

#include "nr.h"
#include "nrutil.h"

#include "constants.h"
#include "joint_ekf_R.h"

#include <stdio.h>
#include <stdlib.h>

#include "gaussj.c"
#include "nrutil.c"
#include "matrix_operations.c"
#include "comp_ss_jacobi_R.c"

/* Small number */
#define MU    (1.0e-6F)

/* Model dimensions */
#define INPUTS          1 // No. of known inputs (u)
#define M_STATES        5 // No. of motor + Butterworth filter states (i, if', if, omega, theta)
#define U_INPUTS        1 // No. of unknown inputs (T_L)
#define U_BIAS          1 // No. of unknown biases (i_b)
#define P_STATES        1 // No. of parameters to estimate (R)
#define MEASUREMENTS	2 // No. of physical measurements (i, theta)

#define NX      (M_STATES)
#define NXX 	(M_STATES + U_INPUTS)
#define NXXX	(M_STATES + U_INPUTS + P_STATES)
#define NXXXX	(M_STATES + U_INPUTS + P_STATES + U_BIAS)
#define NU      (INPUTS)
#define NY      (MEASUREMENTS)

#define WARNING     0
#define DEBUG       0

int INIT_SUCCESS = 0; // Set upon successful initialization, to avoid segmentation fault mdlTerminate().

/* Error handling
 * --------------
 *
 * You should use the following technique to report errors encountered within
 * an S-function:
 *
 *       ssSetErrorStatus(S,"Error encountered due to ...");
 *       return;
 *
 * Note that the 2nd argument to ssSetErrorStatus must be persistent memory.
 * It cannot be a local variable. For example the following will cause
 * unpredictable errors:
 *
 *      mdlOutputs()
 *      {
 *         char msg[256];         {ILLEGAL: to fix use "static char msg[256];"}
 *         sprintf(msg,"Error due to %s", string);
 *         ssSetErrorStatus(S,msg);
 *         return;
 *      }
 *
 * See matlabroot/simulink/src/sfuntmpl_doc.c for more details.
 */

/* Function: mdlInitializeSizes ===============================================
 * Abstract:
 *    The sizes information is used by Simulink to determine the S-function
 *    block's characteristics (number of inputs, outputs, states, etc.).
 */
static void mdlInitializeSizes(SimStruct *S)
{
    /* See sfuntmpl_doc.c for more details on the macros below */
    
    ssSetNumSFcnParams(S, 2);  /* Number of expected parameters */
    if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) {
        /* Return if number of expected != number of actual parameters */
        return;
    }
    
    ssSetNumContStates(S, 0);
    ssSetNumDiscStates(S, NXXXX);
    
    if (!ssSetNumInputPorts(S, 2)) return;
    ssSetInputPortWidth(S, 0, NU);
    ssSetInputPortWidth(S, 1, NY);
    ssSetInputPortRequiredContiguous(S, 0, true); /*direct input signal access*/
    ssSetInputPortRequiredContiguous(S, 1, true);
    /*
     * Set direct feedthrough flag (1=yes, 0=no).
     * A port has direct feedthrough if the input is used in either
     * the mdlOutputs or mdlGetTimeOfNextVarHit functions.
     * See matlabroot/simulink/src/sfuntmpl_directfeed.txt.
     */
    ssSetInputPortDirectFeedThrough(S, 0, 1);
    ssSetInputPortDirectFeedThrough(S, 1, 1);
    
    if (!ssSetNumOutputPorts(S, 1)) return;
    ssSetOutputPortWidth(S, 0, NXXXX);
    
    ssSetNumSampleTimes(S, 1);
    ssSetNumRWork(S, 0);
    ssSetNumIWork(S, 0);
    ssSetNumPWork(S, 13);
    ssSetNumModes(S, 0);
    ssSetNumNonsampledZCs(S, 0);
    
    ssSetOptions(S, SS_OPTION_EXCEPTION_FREE_CODE);
}

/* Function: mdlInitializeSampleTimes =========================================
 * Abstract:
 *    This function is used to specify the sample time(s) for your
 *    S-function. You must register the same number of sample times as
 *    specified in ssSetNumSampleTimes.
 */
static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S, 0, Ts); // 1 ms
    ssSetOffsetTime(S, 0, 0.0);
}

#define MDL_INITIALIZE_CONDITIONS   /* Change to #undef to remove function */
#if defined(MDL_INITIALIZE_CONDITIONS)
  /* Function: mdlInitializeConditions ========================================
   * Abstract:
   *    In this function, you should initialize the continuous and discrete
   *    states for your S-function block.  The initial states are placed
   *    in the state vector, ssGetContStates(S) or ssGetRealDiscStates(S).
   *    You can also perform any other initialization activities that your
   *    S-function may require. Note, this routine will be called at the
   *    start of simulation and if it is present in an enabled subsystem
   *    configured to reset states, it will be call when the enabled subsystem
   *    restarts execution to reset the states.
   */
static void mdlInitializeConditions(SimStruct *S)
{
    int i;
    
    const real_T *x0;
    const real_T *q0;
    
    real_T *x;
    int_T nDStates;
    int_T n;
    
    float **T;
    float **Cd, **CdT;
    float **K, **P, **Q, **R, **I;
	
    x = ssGetRealDiscStates(S);
    nDStates = ssGetNumDiscStates(S);
    
    x0 = mxGetPr(ssGetSFcnParam(S,0));
    n = mxGetNumberOfElements(ssGetSFcnParam(S,0));
    
    if (nDStates != n) {
        ssSetErrorStatus(S,"Error encountered due wrong number of initial values for the state vector.");
        return;
    }
    for (i = 0; i < n; i++) {
    	x[i] = x0[i]; // Initialize states.
    }
    
    ssSetPWorkValue(S, 0, matrix(1, NXX, 1, NXX)); // Ad, state matrix
    ssSetPWorkValue(S, 1, matrix(1, NXX, 1, NU));  // Bd, input matrix
    ssSetPWorkValue(S, 2, matrix(1, NXX, 1, NXX)); // AdR, linearization with respect to resistance
    ssSetPWorkValue(S, 3, matrix(1, NXX, 1, NU));  // BdR
    
    Cd  = matrix(1, NY, 1, NXXXX);
    CdT = matrix(1, NXXXX, 1, NY);
    
    init_zeros(Cd, NY, NXXXX);
    Cd[1][3] = 1.0F; // Current output of Butterworth filter
    Cd[1][8] = 1.0F; // Current measurement bias
    Cd[2][5] = 1.0F; // Position (encoder)
    ssSetPWorkValue(S, 4, Cd);  // Cd, measurement matrix
    
    matrix_transpose(Cd, CdT, NY, NXXXX);
    ssSetPWorkValue(S, 5, CdT); // CdT, measurement matrix transpose
    
    K = matrix(1, NXXXX, 1, NY);
    init_zeros(K, NXXXX, NY);
    ssSetPWorkValue(S, 6, K); // K, Kalman gain matrix
    
    P = matrix(1, NXXXX, 1, NXXXX);
    init_ident(P, NXXXX);
    ssSetPWorkValue(S, 7, P); // P, Error variance
    
    Q = matrix(1, NXXXX, 1, NXXXX);
    init_zeros(Q, NXXXX, NXXXX);
    
    q0 = mxGetPr(ssGetSFcnParam(S,1));
    n = mxGetNumberOfElements(ssGetSFcnParam(S,1));
    
    if (nDStates != n) {
        ssSetErrorStatus(S,"Error encountered due wrong number of initializing values for the process noise matrix.");
        return;
    }
    for (i = 0; i < n; i++) {
        Q[i+1][i+1] = (float)q0[i]; // Initialize diagonal matrix elements.
    }
    ssSetPWorkValue(S, 8, Q); // Q, Process noise variance
    
    printf("\nQ\n");  print_matrix(Q, NXXXX, NXXXX);
    
    R = matrix(1, NY, 1, NY);
    init_zeros(R, NY, NY);
    R[1][1] = (float)pow(0.002F,2); // Current measurement noise variance (at omega = 0)
    R[2][2] = (float)pow(M_PI/1000.0F,2)/12.0F; // Quantizer (encoder) noise variance, assuming uniform distribution
    ssSetPWorkValue(S, 9, R); // R, Measurement noise variance
    
    I = matrix(1, NXXXX, 1, NXXXX);
    init_ident(I, NXXXX);
    ssSetPWorkValue(S, 10, I); // I, NXXX x NXXX identity matrix
    
    // Augmented matrices
    T = matrix(1, NXXXX, 1, NXXXX);
    init_ident(T, NXXXX);
    ssSetPWorkValue(S, 11, T); // Ad_, state matrix
    
    T = matrix(1, NXXXX, 1, NY);
    init_zeros(T, NXXXX, NY);
    ssSetPWorkValue(S, 12, T); // Bd_, input matrix
    
    #if (DEBUG)
    printf("\nCd\n"); print_matrix(Cd, NY, NXXXX);
    printf("\nCdT\n"); print_matrix(CdT, NXXXX, NY);
    printf("\nK\n"); print_matrix(K, NXXXX, NY);
    printf("\nP\n");  print_matrix(P, NXXXX, NXXXX);
    printf("\nQ\n"); print_matrix(Q, NXXXX, NXXXX);
    printf("\nR\n"); print_matrix(R, NY, NY);
    #endif
    
    INIT_SUCCESS = 1;
}
#endif /* MDL_INITIALIZE_CONDITIONS */

#undef MDL_START  /* Change to #undef to remove function */
#if defined(MDL_START)
  /* Function: mdlStart =======================================================
   * Abstract:
   *    This function is called once at start of model execution. If you
   *    have states that should be initialized once, this is the place
   *    to do it.
   */
static void mdlStart(SimStruct *S)
{
    
}
#endif /*  MDL_START */

/* Function: mdlOutputs =======================================================
 * Abstract:
 *    In this function, you compute the outputs of your S-function
 *    block. Generally outputs are placed in the output vector, ssGetY(S).
 */
static void mdlOutputs(SimStruct *S, int_T tid)
{
    int i;
    
    const real_T *uu, *yy;
    real_T *zz, *xx;
    float **u, **y, **x;
    
    float **Ad,  **Bd,  **AdR,  **BdR;
    float **Ad_, **Bd_;
    float **Cd, **CdT;
    float **K, **P, **Q, **R, **I;
    
    float **dR;
    
    float R_hat;
    
    float check;
    
    float **T1, **T2, **T3, **t1, **t2, **err;
    
	uu = (const real_T *)ssGetInputPortSignal(S,0); // Voltage input
    u = matrix(1, NU, 1, 1);
    for (i = 1; i <= NU; i++) {
            u[i][1] = (float)uu[i-1];
    }
    #if (DEBUG)
    printf("\nu\n"); print_matrix(u, NU, 1);
    #endif
    
    yy = (const real_T *)ssGetInputPortSignal(S,1); // Measurements
    y = matrix(1, NY, 1, 1);
    for (i = 1; i <= NY; i++) {
            y[i][1] = (float)yy[i-1];
    }
    #if (DEBUG)
    printf("\ny\n"); print_matrix(y, NY, 1);
    #endif
    
    zz = ssGetOutputPortSignal(S,0); // Kalman filter output
    
    xx = ssGetRealDiscStates(S); // Kalman filter states
    x = matrix(1, NXXXX, 1, 1);
    for (i = 1; i <= NXXXX; i++) {
            x[i][1] = (float)xx[i-1];
    }
    #if (DEBUG)
    printf("\nx\n"); print_matrix(x, NXXXX, 1);
    #endif
    
    Ad  = ssGetPWorkValue(S, 0);
    Bd  = ssGetPWorkValue(S, 1);
    AdR = ssGetPWorkValue(S, 2);
    BdR = ssGetPWorkValue(S, 3);
    
    Cd  = ssGetPWorkValue(S, 4);
    CdT = ssGetPWorkValue(S, 5);
    
    K = ssGetPWorkValue(S, 6); 
    P = ssGetPWorkValue(S, 7); 
    Q = ssGetPWorkValue(S, 8); 
    R = ssGetPWorkValue(S, 9);
    
    I = ssGetPWorkValue(S, 10);
    
    Ad_ = ssGetPWorkValue(S, 11);
    Bd_ = ssGetPWorkValue(S, 12);
    
    // Update measurement noise for current, velocity dependent
    R[1][1] = (float)pow((0.00025*(float)fabs(x[4][1]) + 0.002F), 2);
    
    T1 = matrix(1, NXXXX, 1, NY); T2 = matrix(1, NY, 1, NY);
    // Compute Kalman gain: K = P_*Cd'*(Cd*P_*Cd'+R)^-1
    matrix_multiply(P, CdT, T1, NXXXX, NXXXX, NY); // T1 = P_*Cd'
    matrix_multiply(Cd, T1, T2, NY, NXXXX, NY);    // T2 = Cd*P_*Cd'
    matrix_addition(T2, R, T2, NY, NY);            // T2 = Cd*P_*Cd'+R
    gaussj(T2, NY, NULL, 0);                       // T2 = (Cd*P_*Cd'+R)^-1
    matrix_multiply(T1, T2, K, NXXXX, NY, NY);     // K  = P_*Cd'*(Cd*P_*Cd'+R)^-1
    #if (DEBUG)
    printf("\nK\n"); print_matrix(K, NXXXX, NY);
    #endif
    free_matrix(T1, 1, NXXXX, 1, NY); free_matrix(T2, 1, NY, 1, NY);
    
    t1 = matrix(1, NY, 1, 1); t2 = matrix(1, NXXXX, 1, 1); err = matrix(1, NY, 1, 1);
    // Update estimate with measurement (filtering): x_hat = x_hat_ + K*(y - Cd*x_hat_)
    matrix_multiply(Cd, x, t1, NY, NXXXX, 1);  // t1 = Cd*x_hat_
    matrix_subtraction(y, t1, err, NY, 1);     // err = y - Cd*x_hat_
    matrix_multiply(K, err, t2, NXXXX, NY, 1); // t2 = K*(y - Cd*x_hat_)
    matrix_addition(x, t2, x, NXXXX, 1);       // x_hat = x_hat_ + K*(y - D*x_hat_)
    #if (DEBUG)
    printf("\nx\n"); print_matrix(x, NXXXX, 1);
    #endif
    free_matrix(t1, 1, NY, 1, 1); free_matrix(t2, 1, NXXXX, 1, 1); free_matrix(err, 1, NY, 1, 1);
    
    T1 = matrix(1, NXXXX, 1, NXXXX); T2 = matrix(1, NXXXX, 1, NXXXX);
    // Compute error covariance: P = (I - K*Cd)*P_
    matrix_multiply(K, Cd, T1, NXXXX, NY, NXXXX);    // T1 = K*Cd
    matrix_subtraction(I, T1, T1, NXXXX, NXXXX);     // T1 = I - K*Cd
    matrix_multiply(T1, P, T2, NXXXX, NXXXX, NXXXX); // T2 = (I - K*Cd)*P_
    copy_matrix(T2, P, 1, NXXXX, 1, NXXXX);          // P = T2
    #if (DEBUG)
    printf("\nP\n"); print_matrix(P, NXXXX, NXXXX);
    #endif
    
    // Project ahead (prediction): x_hat_ = f(x_hat, u), P_ = Phi*P*Phi' + Q
    R_hat = x[NXX+1][1]; // Indexing from one
    // Sanity check: avoid imaginary parts in the eigenvalues and unphysical parameter estimates.
    if ( R_hat < 2.0F ) {
        x[NXX+1][1] = R_hat = 2.0F;
        #if (WARNING)
        printf("R_hat seem unphysical, resetting to %e\n", x[NXX+1][1]);
        #endif
    }
    check = (float)sqrt((4*L_*K_E_*K_T_)/J_); // R_hat
    if (DEBUG) printf("check %e\n", check);
    if ( R_hat <= check ) {
        x[NXX+1][1] = R_hat = check + MU;
        #if (WARNING)
        printf("R_hat to small, resetting to %e\n", x[NXX+1][1]);
        #endif
	}
    
    // Computing state space and Jacobians for current parameter estimates
    comp_ss_jacobi(Ad, Bd, AdR, BdR, R_hat, Ts_);
    #if (DEBUG)
    printf("\nAd\n"); print_matrix(Ad, NXX, NXX);
    printf("\nBd\n"); print_matrix(Bd, NXX, NU);
    printf("\nAdR\n"); print_matrix(AdR, NXX, NXX);
    printf("\nBdR\n"); print_matrix(AdR, NXX, NU);
    #endif
    
    // Augmenting state matrix with T_L, R
    init_ident(Ad_, NXXXX);
    copy_matrix(Ad, Ad_, 1, NXX, 1, NXX);
    #if (DEBUG)
    printf("\nAd_\n"); print_matrix(Ad_, NXXXX, NXXXX);
    #endif
    
    copy_matrix(Bd, Bd_, 1, NXX, 1, NU);
    #if (DEBUG)
    printf("\nBd_\n"); print_matrix(Bd_, NXX, NU);
    #endif
    
    t1  = matrix(1, NXXXX, 1, 1); t2  = matrix(1, NXXXX, 1, 1);
    // x_hat_ = f(x_hat, u)
    matrix_multiply(Ad_, x, t1, NXXXX, NXXXX, 1);
    matrix_multiply(Bd_, u, t2, NXXXX, NU, 1);
    matrix_addition(t1, t2, x, NXXXX, 1); // x_hat_
    #if (DEBUG)
    printf("\nx\n"); print_matrix(x, NXXXX, 1);
    #endif
    free_matrix(t1, 1, NXXXX, 1, 1); free_matrix(t2, 1, NXXXX, 1, 1);
    
    dR = matrix(1, NXX, 1, 1);
    t1 = matrix(1, NXX, 1, 1); t2  = matrix(1, NXX, 1, 1);
    // Computing linearized elements for given operating point
    matrix_multiply(AdR, x, t1, NXX, NXX, 1);
    matrix_multiply(BdR, u, t2, NXX, NU, 1);
    matrix_addition(t1, t2, dR, NXX, 1);
    for (i = 1; i <= NXX; i++) {
        Ad_[i][NXX+1] = dR[i][1]; // Building linearized state matrix
    }
    #if (DEBUG)
    printf("\nAd_\n"); print_matrix(Ad_, NXXXX, NXXXX);
    #endif
    free_matrix(dR, 1, NXX, 1, 1);
    free_matrix(t1, 1, NXX, 1, 1); free_matrix(t2, 1, NXX, 1, 1);
    
    T3 = matrix(1, NXXXX, 1, NXXXX);
    
    // P_ = Phi*P*Phi' + Q
    matrix_multiply(Ad_, P, T1, NXXXX, NXXXX, NXXXX); // T1 = Phi*P
    matrix_transpose(Ad_, T2, NXXXX, NXXXX);          // T2 = Phi'
    matrix_multiply(T1, T2, T3, NXXXX, NXXXX, NXXXX); // T3 = Phi*P*Phi'
    matrix_addition(T3, Q, P, NXXXX, NXXXX);          // P_ = Phi*P*Phi' + Q
    #if (DEBUG)
    printf("\nP_\n"); print_matrix(P, NXXXX, NXXXX);
    #endif
    free_matrix(T1, 1, NXXXX, 1, NXXXX); free_matrix(T2, 1, NXXXX, 1, NXXXX); free_matrix(T3, 1, NXXXX, 1, NXXXX);
    
    // Kalman filter output
    for (i = 1; i <= NXXXX; i++) {
            zz[i-1] = (real_T)x[i][1];
    }

    // Kalman filter states
    for (i = 1; i <= NXXXX; i++) {
            xx[i-1] = (real_T)x[i][1];
    }

    free_matrix(u, 1, NU, 1, 1); free_matrix(y, 1, NY, 1, 1); free_matrix(x, 1, NXXXX, 1, 1);
}



#undef MDL_UPDATE  /* Change to #undef to remove function */
#if defined(MDL_UPDATE)
  /* Function: mdlUpdate ======================================================
   * Abstract:
   *    This function is called once for every major integration time step.
   *    Discrete states are typically updated here, but this function is useful
   *    for performing any tasks that should only take place once per
   *    integration step.
   */
static void mdlUpdate(SimStruct *S, int_T tid)
{
}
#endif /* MDL_UPDATE */



#undef MDL_DERIVATIVES  /* Change to #undef to remove function */
#if defined(MDL_DERIVATIVES)
  /* Function: mdlDerivatives =================================================
   * Abstract:
   *    In this function, you compute the S-function block's derivatives.
   *    The derivatives are placed in the derivative vector, ssGetdX(S).
   */
static void mdlDerivatives(SimStruct *S)
{
}
#endif /* MDL_DERIVATIVES */



/* Function: mdlTerminate =====================================================
 * Abstract:
 *    In this function, you should perform any actions that are necessary
 *    at the termination of a simulation.  For example, if memory was
 *    allocated in mdlStart, this is the place to free it.
 */
static void mdlTerminate(SimStruct *S)
{
    if (INIT_SUCCESS) {
        
        float **Ad,  **Bd,  **AdR,  **BdR;
        float **Ad_, **Bd_;
        float **Cd, **CdT;
        float **K, **P, **Q, **R, **I;
		
        Ad    = ssGetPWorkValue(S, 0);
        free_matrix(Ad, 1, NXX, 1, NXX);
        Bd    = ssGetPWorkValue(S, 1);
        free_matrix(Bd, 1, NXX, 1, NU);
        AdR   = ssGetPWorkValue(S, 2);
        free_matrix(AdR, 1, NXX, 1, NXX);
        BdR   = ssGetPWorkValue(S, 3);
        free_matrix(BdR, 1, NXX, 1, NU);
        
        Cd  = ssGetPWorkValue(S, 4);
        free_matrix(Cd, 1, NY, 1, NXXXX);
        CdT = ssGetPWorkValue(S, 5);
        free_matrix(CdT, 1, NXXXX, 1, NY);
        
        K = ssGetPWorkValue(S, 6);
        free_matrix(K, 1, NXXXX, 1, NY);
        
        P = ssGetPWorkValue(S, 7);
        printf("\nP\n"); print_matrix(P, NXXXX, NXXXX); // Print final error covariance
        free_matrix(P, 1, NXXXX, 1, NXXXX);
        
        Q = ssGetPWorkValue(S, 8);
        free_matrix(Q, 1, NXXXX, 1, NXXXX);
        R = ssGetPWorkValue(S, 9);
        free_matrix(R, 1, NY, 1, NY);
        
        I = ssGetPWorkValue(S, 10);
        free_matrix(I, 1, NXXXX, 1, NXXXX);
        
        Ad_ = ssGetPWorkValue(S, 11);
        free_matrix(Ad_, 1, NXXXX, 1, NXXXX);
        Bd_ = ssGetPWorkValue(S, 12);
        free_matrix(Bd_, 1, NXXXX, 1, NU);
        
    }
}



/*======================================================*
 * See sfuntmpl_doc.c for the optional S-function methods *
 *======================================================*/

/*=============================*
 * Required S-function trailer *
 *=============================*/

#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
