% Outputs discrete and linearized state and input matrices for combined
% DC motor and Butterworth filter model.
% Arnfinn Aas Eielsen, 2007
function [Phi, Delta, PhiD, DeltaD] = compute_discrete_ss_and_jacobi(R, J, K_E, K_T, T)

if nargin ~= 5
    error('Wrong number of arguments...')
end

Phi = zeros(5,5);
Delta = zeros(5,2);
PhiD = zeros(5,5,3);
DeltaD = zeros(5,2,3);

% Data for DC motor (Maxon RE 36.118800)
L = 0.487e-3;%mH -> H
%J = 6.78e-6;%g cm^2 -> kg m^2
%R = 2.74;%Omega
%K_T = 56.6e-3; %mNm/A -> Nm/A
%K_E = K_T; %V/(rad/s)

% Sampling time
%T = 1e-3; %s

% Butterworth filter cut-off frequency
wc = 2*pi*100; %rad/s

d = J^2*R^2 - 4*J*L*K_T*K_E;

l1 = (-J*R-sqrt(d))/(2*J*L);
l2 = (-J*R+sqrt(d))/(2*J*L);
e1 = exp(l1*T);
e2 = exp(l2*T);

lx = wc*sqrt(2)/2;
ex = exp(-lx*T);
cosx = cos(lx*T);
sinx = sin(lx*T);
w = l1 - l2;

% State matrix
Phi(1,1) = (l1*e1 - l2*e2)/w;
Phi(1,2) = 0;
Phi(1,3) = 0;
Phi(1,4) = (K_E/L)*(e2 - e1)/w;
Phi(1,5) = 0;

Phi(2,1) = compute_f(3, wc, l1, l2, -lx, lx, T);
Phi(2,2) = ( cosx - sinx )*ex;
Phi(2,3) = -wc*( 1/lx*sinx )*ex;
Phi(2,4) = compute_f(2, -(wc*K_E)/L, l1, l2, -lx, lx, T);
Phi(2,5) = 0;

Phi(3,1) = compute_f(2, pow(wc,2), l1, l2, -lx, lx, T);
Phi(3,2) = -Phi(2,3);
Phi(3,3) = ( sinx + cosx )*ex;
Phi(3,4) = compute_f(1, -(pow(wc,2)*K_E)/L, l1, l2, -lx, lx, T);
Phi(3,5) = 0;

Phi(4,1) = ((K_T/J)*(e1 - e2))/w;
Phi(4,2) = 0;
Phi(4,3) = 0;
Phi(4,4) = (l1*e2 - l2*e1)/w;
Phi(4,5) = 0;

Phi(5,1) = ((K_T/J)*((1/l1)*(e1 - 1) - (1/l2)*(e2 - 1)))/w;
Phi(5,2) = 0;
Phi(5,3) = 0;
Phi(5,4) = (((l1/l2)*(e2 - 1) - (l2/l1)*(e1 - 1)))/w;
Phi(5,5) = 1;

% Input matrix
Delta(1,1) = J/(K_T*L)*Phi(4,1);
Delta(2,1) = compute_f(2, wc/L, l1, l2, -lx, lx, T);
Delta(3,1) = compute_f(1, pow(wc,2)/L, l1, l2, -lx, lx, T);
Delta(4,1) = 1/L*Phi(5,1);
Delta(5,1) = (K_T/(J*L)*( 1/pow(l1,2)*(e1 - 1) - 1/pow(l2,2)*(e2 - 1) + (1/l2 - 1/l1)*T))/w;

Delta(1,2) = K_E/K_T*Delta(4,1);
Delta(2,2) = compute_f(1, (wc*K_E)/(J*L), l1, l2, -lx, lx, T);
Delta(3,2) = compute_f(0, (pow(wc,2)*K_E)/(J*L), l1, l2, -lx, lx, T);
Delta(4,2) = -1/J*Phi(5,4);
Delta(5,2) = ((-1/J)*( l1/pow(l2,2)*(e2 - 1) - l2/pow(l1,2)*(e1 - 1) + (l2/l1 - l1/l2)*T))/w;

% Linearization
k = [1, K_E/L, wc, -(wc*K_E)/L, pow(wc,2), -(pow(wc,2)*K_E)/L, K_T/J, 1, K_T/J, 1, ...
        J/(K_T*L), wc/L, pow(wc,2)/L, 1/L, K_T/(J*L), K_E/(K_T*L), (wc*K_E)/(J*L), (pow(wc,2)*K_E)/(J*L), -1/J, -1/J];

for i=1:5
    switch i
        case 1 % Resistance - R
            l1D = -1/(2*L)*(1 + (J*R)/sqrt(d));
            l2D = -1/(2*L)*(1 - (J*R)/sqrt(d));
            
            kD = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ...
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
        case 2 % Speed constant - K_E
            l1D = K_T/sqrt(d);
            l2D = -K_T/sqrt(d);
            
            kD = [0, 1/L, 0, -wc/L, 0, -pow(wc,2)/L, 0, 0, 0, 0, ...
                    0, 0, 0, 0, 0, 1/(K_T*L), wc/(J*L), pow(wc,2)/(J*L), 0, 0];
        case 3 % Torque constant - K_T
            l1D = K_E/sqrt(d);
            l2D = -K_E/sqrt(d);
            
            kD = [0, 0, 0, 0, 0, 0, 1/J, 0, 1/J, 0, ...
                    -J/(pow(K_T,2)*L), 0, 0, 0, 1/(J*L), -K_E/(pow(K_T,2)*L), 0, 0, 0, 0];
        case 4 % Speed/torque constant - K_E = K_T = K
            if (K_E ~= K_T)
                error('K_T and K_E are not equal.');
            end
            K = K_E;
            l1D = (2*K)/sqrt(d);
            l2D = -(2*K)/sqrt(d);
            
            kD = [0, 1/L, 0, -wc/L, 0, -pow(wc,2)/L, 1/J, 0, 1/J, 0, ...
                    -J/(pow(K,2)*L), 0, 0, 0, 1/(J*L), 0, wc/(J*L), pow(wc,2)/(J*L), 0, 0];
        case 5 % Inertia - J
            l1D = -(K_T*K_E)/(J*sqrt(d));
            l2D = (K_T*K_E)/(J*sqrt(d));
            
            kD = [0, 0, 0, 0, 0, 0, -K/pow(J,2), 0, -K/pow(J,2), 0, ...
                    1/(K*L), 0, 0, 0, -K/(pow(J,2)*L), 0, -(wc*K)/(pow(J,2)*L), -(pow(wc,2)*K)/(pow(J,2)*L), 1/pow(J,2), 1/pow(J,2)];
    end

    e1D = l1D*T*e1;
    e2D = l2D*T*e2;
    wD = l1D - l2D;
    
    % Linearized state matrix
    PhiD(1,1,i) = comp_lin(k(1), kD(1), w, wD, (l1*e1 - l2*e2), (l1D*e1 + l1*e1D - l2D*e2 - l2*e2D));
    PhiD(1,2,i) = 0;
    PhiD(1,3,i) = 0;
    PhiD(1,4,i) = comp_lin(k(2), kD(2), w, wD, (e2 - e1), (e2D - e1D));
    PhiD(1,5,i) = 0;
    
    PhiD(2,1,i) = compute_fD(3, k(3), kD(3), l1, l1D, l2, l2D, -lx, lx, T);
    PhiD(2,2,i) = 0;
    PhiD(2,3,i) = 0;
    PhiD(2,4,i) = compute_fD(2, k(4), kD(4), l1, l1D, l2, l2D, -lx, lx, T);
    PhiD(2,5,i) = 0;
    
    PhiD(3,1,i) = compute_fD(2, k(5), kD(5), l1, l1D, l2, l2D, -lx, lx, T);
    PhiD(3,2,i) = 0;
    PhiD(3,3,i) = 0;
    PhiD(3,4,i) = compute_fD(1, k(6), kD(6), l1, l1D, l2, l2D, -lx, lx, T);
    PhiD(3,5,i) = 0;
    
    PhiD(4,1,i) = comp_lin(k(7), kD(7), w, wD, (e1 - e2), (e1D - e2D));
    PhiD(4,2,i) = 0;
    PhiD(4,3,i) = 0;
    PhiD(4,4,i) = comp_lin(k(8), kD(8), w, wD, (l1*e2 - l2*e1), (l1D*e2 + l1*e2D - l2D*e1 - l2*e1D));
    PhiD(4,5,i) = 0;
    
    PhiD(5,1,i) = comp_lin(k(9), kD(9), w, wD, ((e1 - 1)/l1 - (e2 - 1)/l2), ...
        (e1D/l1 - (l1D*(e1 - 1))/pow(l1,2) - e2D/l2 + (l2D*(e2 - 1))/pow(l2,2)));
    PhiD(5,2,i) = 0;
    PhiD(5,3,i) = 0;
    PhiD(5,4,i) = comp_lin(k(10), kD(10), w, wD, ((l1*(e2 - 1))/l2 - (l2*(e1 - 1))/l1), ...
        ((l1D*l2 - l1*l2D)/pow(l2,2)*(e2 - 1) + (l1*e2D)/l2 - (l2D*l1 - l2*l1D)/pow(l1,2)*(e1 - 1) - (l2*e1D)/l1));
    PhiD(5,5,i) = 0;
    
    % Linearized input matrix
    DeltaD(1,1,i) = kD(11)*Phi(4,1) + k(11)*PhiD(4,1,i);
    DeltaD(2,1,i) = compute_fD(2, k(12), kD(12), l1, l1D, l2, l2D, -lx, lx, T);
    DeltaD(3,1,i) = compute_fD(1, k(13), kD(13), l1, l1D, l2, l2D, -lx, lx, T);
    DeltaD(4,1,i) = kD(14)*Phi(5,1) + k(14)*PhiD(5,1,i);
    DeltaD(5,1,i) = comp_lin(k(15), kD(15), w, wD, ((e1 - 1)/pow(l1,2) - (e2 - 1)/pow(l2,2) + (1/l2 - 1/l1)*T), ...
                        ( -(2*l1*l1D)/pow(l1,4)*(e1 - 1) + e1D/pow(l1,2) + (2*l2*l2D)/pow(l2,4)*(e2 - 1) - ...
                            e2D/pow(l2,2)) + (l1D/pow(l1,2) - l2D/pow(l2,2))*T );
    
    DeltaD(1,2,i) = kD(16)*Phi(5,1) + k(16)*PhiD(5,1,i);
    DeltaD(2,2,i) = compute_fD(1, k(17), kD(17), l1, l1D, l2, l2D, -lx, lx, T);
    DeltaD(3,2,i) = compute_fD(0, k(18), kD(18), l1, l1D, l2, l2D, -lx, lx, T);
    DeltaD(4,2,i) = kD(19)*Phi(5,4) + k(19)*PhiD(5,4,i);
    DeltaD(5,2,i) = comp_lin(k(20), kD(20), w, wD, (l1/pow(l2,2)*(e2 - 1) - l2/pow(l1,2)*(e1 - 1) + (l2/l1 - l1/l2)*T), ...
        ( (l1D*pow(l2,2) - l1*2*l2*l2D)/pow(l2,4)*(e2 - 1) + (l1*e2D)/pow(l2,2) - ...
            (l2D*pow(l1,2) - l2*2*l1*l1D)/pow(l1,4)*(e1 - 1) - (l2*e1D)/pow(l1,2) + ((l2D*l1 - l2*l1D)/pow(l1,2) - ...
                (l1D*l2 - l1*l2D)/pow(l2,2))*T ));
end

% Computes some of the matrix elements
function f = compute_f(s, K, a, b, alpha, beta, t)
d1 = (2*alpha*a - pow(a,2) - pow(alpha,2) - pow(beta,2));
d2 = (2*alpha*b - pow(b,2) - pow(alpha,2) - pow(beta,2));

A1 = K/(d1*(b - a));
A2 = K/(d2*(a - b));
M = K*(a + b - 2*alpha)/(d1*d2);
N = K*(3*pow(alpha,2) - pow(beta,2) - 2*alpha*(a + b) + a*b)/(d1*d2);

switch s
    case 0
        u = (N + alpha*M)/beta;
        v = M;
        f = A1*(1/a)*(exp(a*t) - 1) + A2*(1/b)*(exp(b*t) - 1) + ...
            ( u*((alpha*sin(beta*t) - beta*cos(beta*t))*exp(alpha*t) + beta) + ...
                v*((alpha*cos(beta*t) + beta*sin(beta*t))*exp(alpha*t) - alpha) )/(pow(alpha,2) + pow(beta,2));
    case 1
        u = (N + alpha*M)/beta;
        v = M;
        f = A1*exp(a*t) + A2*exp(b*t) + ( u*sin(beta*t) + v*cos(beta*t) )*exp(alpha*t);
    case 2
        u = ( alpha*N + (pow(alpha,2) - pow(beta,2))*M )/beta;
        v = N + 2*alpha*M;
        f = a*A1*exp(a*t) + b*A2*exp(b*t) + ( u*sin(beta*t) + v*cos(beta*t) )*exp(alpha*t);
    case 3
        u = ( (pow(alpha,2) - pow(beta,2))*N + (pow(alpha,3) - 3*alpha*pow(beta,2))*M )/beta;
        v = 2*alpha*N + (3*pow(alpha,2) - pow(beta,2))*M;
        f = pow(a,2)*A1*exp(a*t) + pow(b,2)*A2*exp(b*t) + ( u*sin(beta*t) + v*cos(beta*t) )*exp(alpha*t);
end

% Computes some of the matrix elements
function fD = compute_fD(s, K, KD, a, aD, b, bD, alpha, beta, t)
d1 = (2*alpha*a - pow(a,2) - pow(alpha,2) - pow(beta,2));
d2 = (2*alpha*b - pow(b,2) - pow(alpha,2) - pow(beta,2));
d1D = 2*alpha*aD - 2*a*aD;
d2D = 2*alpha*bD - 2*b*bD;

A1 = K/(d1*(b - a));
A2 = K/(d2*(a - b));
A1D = (KD*d1*(b - a) - K*(d1D*(b - a) + d1*(bD - aD)))/pow(d1*(b - a),2);
A2D = (KD*d2*(a - b) - K*(d2D*(a - b) + d2*(aD - bD)))/pow(d2*(b - a),2);
MD = ((KD*(a + b - 2*alpha) + K*(aD + bD))*d1*d2 - K*(a + b - 2*alpha)*(d1D*d2 + d1*d2D))/pow(d1*d2,2);
ND = ((KD*(3*pow(alpha,2) - pow(beta,2) - 2*alpha*(a + b) + a*b) + K*(aD*b + a*bD - 2*alpha*(aD + bD)))*d1*d2 - ...
    K*(3*pow(alpha,2) - pow(beta,2) - 2*alpha*(a + b) + a*b)*(d1D*d2 + d1*d2D))/pow(d1*d2,2);

switch s
    case 0
        uD = (ND + alpha*MD)/beta;
        vD = MD;
        fD = (A1D*(exp(a*t) - 1))/a + (A1*(aD*t*exp(a*t)*a - (exp(a*t) - 1)*aD))/pow(a,2) + ...
            (A2D*(exp(b*t) - 1))/b + (A2*(bD*t*exp(b*t)*b - (exp(b*t) - 1)*bD))/pow(b,2) + ...
                ( uD*((alpha*sin(beta*t) - beta*cos(beta*t))*exp(alpha*t) + beta) + ...
                    vD*((alpha*cos(beta*t) + beta*sin(beta*t))*exp(alpha*t) - alpha) )/(pow(alpha,2) + pow(beta,2));
    case 1
        uD = (ND + alpha*MD)/beta;
        vD = MD;
        fD = A1D*exp(a*t) + A1*aD*t*exp(a*t) + A2D*exp(b*t) + A2*bD*t*exp(b*t) + ( uD*sin(beta*t) + vD*cos(beta*t) )*exp(alpha*t);
    case 2
        uD = ( alpha*ND + (pow(alpha,2) - pow(beta,2))*MD )/beta;
        vD = ND + 2*alpha*MD;
        fD = aD*A1*exp(a*t) + a*A1D*exp(a*t) + a*A1*aD*t*exp(a*t) + bD*A2*exp(b*t) + b*A2D*exp(b*t) + b*A2*bD*t*exp(b*t) + ...
            ( uD*sin(beta*t) + vD*cos(beta*t) )*exp(alpha*t);
    case 3
        uD = ( (pow(alpha,2) - pow(beta,2))*ND + (pow(alpha,3) - 3*alpha*pow(beta,2))*MD )/beta;
        vD = 2*alpha*ND + (3*pow(alpha,2) - pow(beta,2))*MD;
        fD = 2*a*aD*A1*exp(a*t) + pow(a,2)*A1D*exp(a*t) + pow(a,2)*A1*aD*t*exp(a*t) + ...
            2*b*bD*A2*exp(b*t) + pow(b,2)*A2D*exp(b*t) + pow(b,2)*A2*bD*t*exp(b*t) + ...
                ( uD*sin(beta*t) + vD*cos(beta*t) )*exp(alpha*t);
end

% Computes some of the matrix elements
function l = comp_lin(k, kD, w, wD, g, gD)
l = k*(gD*w - g*wD)/pow(w,2) + (kD*g)/w;

% C style exponentiation function
function r = pow(a, b)
r = a^b;
