% Observability analysis, discrete system.
% Arnfinn Aas Eielsen, 2007

% Data for DC motor (Maxon RE 36.118800)
R = 2.74; %Ohm
J = 6.78e-6; %g cm^2 -> kg m^2
K_T = 56.6e-3; %mNm/A -> Nm/A
K_E = K_T; %V/(rad/s)

% Sampling time
T = 1e-3; %s

% Compute discrete state space matrices and their Jacobians
[Ad, Bd, AdD, BdD] = compute_discrete_ss_and_jacobi(R, J, K_E, K_T, T);

x_0 = 1e1*rand(6,1); % Generate some random operating point
u_0 = 1e1*rand;

% Augment state matrix with T_L
Ad = [[Ad; zeros(1,5)], [Bd(:,2); 1]];
% Remove T_L from input matrix
Bd = [Bd(:,1); 0];

if 0 % Turn on/off velocity measurement
    Cd = [0 0 1 0 0 0; 0 0 0 1 0 0; 0 0 0 0 1 0]; % Velocity measurement
else
    Cd = [0 0 1 0 0 0; 0 0 0 0 1 0];
end

% Augment linearized state matrices with T_L
AdR = [[AdD(:,:,1); zeros(1,5)], [BdD(:,2,1); 0]];
BdR = [BdD(:,1,1); 0];

AdK_E = [[AdD(:,:,2); zeros(1,5)], [BdD(:,2,2); 0]];
BdK_E = [BdD(:,1,2); 0];

AdK_T = [[AdD(:,:,3); zeros(1,5)], [BdD(:,2,3); 0]];
BdK_T = [BdD(:,1,3); 0];

AdK = [[AdD(:,:,4); zeros(1,5)], [BdD(:,2,4); 0]];
BdK = [BdD(:,1,4); 0];

AdJ = [[AdD(:,:,5); zeros(1,5)], [BdD(:,2,5); 0]];
BdJ = [BdD(:,1,5); 0];

% Compute linearized elements for given operating point
dR   = AdR*x_0 + BdR*u_0;
dK_T = AdK_T*x_0 + BdK_T*u_0;
dK_E = AdK_E*x_0 + BdK_E*u_0;
dK = AdK*x_0 + BdK*u_0;
dJ = AdJ*x_0 + BdJ*u_0;

% Add R
Ad_{1} = [[Ad; zeros(1,6)], [dR; eye(1)]];
Bd_{1} = [Bd; zeros(1,1)];
Cd_{1} = [Cd, zeros(size(Cd,1),1)];

% Add R, K_T
Ad_{2} = [[Ad; zeros(2,6)], [dR, dK_T; eye(2)]];
Bd_{2} = [Bd; zeros(2,1)];
Cd_{2} = [Cd, zeros(size(Cd,1),2)];

% Add R, K_E
Ad_{3} = [[Ad; zeros(2,6)], [dR, dK_E; eye(2)]];
Bd_{3} = [Bd; zeros(2,1)];
Cd_{3} = [Cd, zeros(size(Cd,1),2)];

% Add R, K_T, K_E
Ad_{4} = [[Ad; zeros(3,6)], [dR, dK_T, dK_E; eye(3)]];
Bd_{4} = [Bd; zeros(3,1)];
Cd_{4} = [Cd, zeros(size(Cd,1),3)];

% Add R, K_T, K_E, i_b
Ad_{5} = [[Ad; zeros(4,6)], [dR, dK_T, dK_E, zeros(6,1); eye(4)]];
Bd_{5} = [Bd; zeros(4,1)];
Cd_{5} = [Cd, zeros(size(Cd,1),4)];
Cd_{5}(1,end) = 1; % Add current bias to measurement

% Add R, K
Ad_{6} = [[Ad; zeros(2,6)], [dR, dK; eye(2)]];
Bd_{6} = [Bd; zeros(2,1)];
Cd_{6} = [Cd, zeros(size(Cd,1),2)];

% Add R, J
Ad_{7} = [[Ad; zeros(2,6)], [dR, dJ; eye(2)]];
Bd_{7} = [Bd; zeros(2,1)];
Cd_{7} = [Cd, zeros(size(Cd,1),2)];

% Add R, K, J
Ad_{8} = [[Ad; zeros(3,6)], [dR, dK, dJ; eye(3)]];
Bd_{8} = [Bd; zeros(3,1)];
Cd_{8} = [Cd, zeros(size(Cd,1),3)];

% Add R, K, i_b
Ad_{9} = [[Ad; zeros(3,6)], [dR, dK, zeros(6,1); eye(3)]];
Bd_{9} = [Bd; zeros(3,1)];
Cd_{9} = [Cd, zeros(size(Cd,1),3)];
Cd_{9}(1,end) = 1; % Add current bias to measurement

% Add R, i_b
Ad_{10} = [[Ad; zeros(2,6)], [dR, zeros(6,1); eye(2)]];
Bd_{10} = [Bd; zeros(2,1)];
Cd_{10} = [Cd, zeros(size(Cd,1),2)];
Cd_{10}(1,end) = 1; % Add current bias to measurement

% Check observability
S = cell(1,length(Ad_));
disp('--------------- Observability ---------------')
for i=1:length(Ad_)
    r(1) = length(Ad_{i});
    
    O = obsv(Ad_{i}, Cd_{i});
    
    r(2) = rank(O); % Method 1

    E = eig(Ad_{i});  % Method 2
    rr = zeros(1,length(E));
    for j=1:length(E)
        rr(j) = rank([Ad_{i} - E(j)*eye(length(Ad_{i})); Cd_{i}]);
    end
    r(3) = min(rr);
    
    [Abar,Bbar,Cbar,T,k] = obsvf(Ad_{i},Bd_{i},Cd_{i},eps('double')); % Method 3
    r(4) = sum(k);
    
    [U{i},S{i},V{i}] = svd(O); % Method 4
    r(5) = sum( sum(S{i} ~= zeros(size(S{i})), 1) );
    
    disp(['System no. ' num2str(i), ', dim.: ' num2str(r(1))])
    disp('Obsv. ranks:')
    disp(r(2:end))
end
