/* Matrix operations - Arnfinn Aas Eielsen, 2007 */
/* A_ij with dimensions (N x M) -> A[M*(i-1) + (j-1)] */

// Multiply two matrices: A (n x o), B (o x m) -> C (n x m)
static void matrixMultiply(long double *A, long double *B, int N, int O, int M, long double *C)
{
    int i, j, k;

    for (i = 0; i < N; i++) { // for all columns of A
        for (j = 0; j < M; j++) { // for all rows of B
            C[M*i+j] = 0.0;
            for (k = 0; k < O; k++) {
                C[M*i+j] = C[M*i+j] + A[O*i+k]*B[M*k+j];
            }
        }
    }
}

// Add two square matrices: A (n x n), B (n x n) -> C (n x n)
static void matrixAddition(long double *A, long double *B, int N, long double *C)
{
    int i, j;

    for (i = 0; i < N; i++) {
        for (j = 0; j < N; j++) {
            C[N*i+j] = A[N*i+j] + B[N*i+j];
        }
    }
}

// Transpose a matrix: A (n x m) -> B (m x n)
static void matrixTranspose(long double *A, int N, int M, long double *B)
{
    int i, j;

    for (i = 0; i < N; i++) {
        for (j = 0; j < M; j++) {
            B[N*j+i] = A[M*i+j];
        }
    }
}

// Gauss elimintaion: C = [A b] (n x n+1) -> x (n x 1)
static void gaussElimintaion(long double *C, long double *x, int N)
{
	int r, k, i, j, p;

	int M = N+1;

	for (k = 0; k < N-1; k++)
	{
		double max_val = 0;
		int r_max = 0;

		int n_row = 0;
		for (r = k; r < N; r++)
		{
			if (C[M*r+k] == 0.0) n_row++;
		}
		if (n_row == N-k) {
			printf("Error: Singular matrix.\n");
			return;
		} else {
			for (r = k; r < N; r++)
			{
				if(abs(C[M*r+k]) > max_val)
				{
					max_val = abs(C[M*r+k]);
					r_max = r;
				}
			}
			if (r_max != k) {
				for (i = 0; i < N+1; i++)
				{
					double temp = C[M*k+i];
					C[M*k+i] = C[M*r_max+i];
					C[M*r_max+i] = temp;
				}
			}
		}

		double Mjk[N];
		for (j = k+1; j < N; j++)
		{
			Mjk[j] = C[M*j+k]/C[M*k+k];
			for (p = k+1; p < N+1; p++)
			{
				C[M*j+p] = C[M*j+p] - Mjk[j]*C[M*k+p];
			}
		}
	}

	if (C[M*(N-1)+(N-1)] == 0) {
		printf("Error: No unique solution.\n");
		return;
	} else {
		x[N-1] = C[M*(N-1)+N]/C[M*(N-1)+(N-1)];

		for (i= N-2; i >= 0; i--)
		{
			double c = 0;
			for (j = i+1; j < N; j++)
			{
				c = c + C[M*i+j]*x[j];
			}
			x[i] = (1/C[M*i+i])*(C[M*i+N] - c);
		}
	}

	return;
}

// Inverse of a square matrix: A (n x n) -> X = A^-1 (n x n)
static void matrixInverse(long double *A, int N, long double *X)
{

    int i, j, k;
    int M;

    M = N+1;

    long double *x = (long double *)malloc(sizeof(long double)*N);
    long double *C = (long double *)malloc(sizeof(long double)*N*M);

    for(k = 0; k < N; k++) {

		printf("k: %d\n", k);

		for(i = 0; i < N; i++) {
			for(j = 0; j < N; j++) {
				C[M*i+j] = A[N*i+j];
				printf("%8.2f  \n", (float)C[M*i+j]);
			}
    	}

        for(i = 0; i < N; i++) {
            C[M*i+N] = 0.0;
            printf("%8.2f  \n", (float)C[M*i+N]);
        }
        C[M*k+N] = 1.0;
        printf("%8.2f  \n", (float)C[M*k+N]);

        gaussElimintaion(C, x, N);

        for(i = 0; i < N; i++) {
            X[N*i+k] = x[i];
            printf("%8.2f  \n", (float)X[N*i+k]);
        }
    }

    free(x);
    free(C);
}

// Copy an array (interpreted as a matrix) to a new memory allocation.
static void matrixCopy(long double *A, int N, int M, long double *B)
{
    int i, j;

    for (i = 0; i < N; i++) {
        for (j = 0; j < M; j++) {
            B[M*i+j] = A[M*i+j];
        }
    }
}

// Format and print an array (interpreted as a matrix) to stdio.
static void printMatrix(long double *C, int N, int M)
{
	int i, j;
	for(i = 0; i < N; i++)
	{
		printf("%2d: ", i);
		for(j = 0; j < M; j++)
		{
			printf("%9.3f ", (float)C[M*i+j]);
		}
		printf("\n");
	}
	return;
}

// Format and print an array (interpreted as a vector) to stdio.
static void printVector(long double *b, int N)
{
	int i;
	for(i = 0; i < N; i++)
	{
		printf("%2d: %9.3f\n", i, (float)b[i]);
	}
	return;
}
